/**
Benchmarking for OpenCL EPGPU library.

Dr. Orion Lawlor, lawlor@alaska.edu, 2011-04-27 (Public Domain)
*/
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include "epgpu.h" /* Orion Lawlor's Expressive Programming for GPU library */
#include "epgpu.cpp" /* <- just compile it here, now.  No compiler/DLL hassles! */

#include "timingCL.h"

#define BENCH_ATOMICS 1 /* atom_inc, atom_add, atom_cmpxch  */
#define BENCH_DOUBLE 1 /* double precision */


void dump_image(const gpu_vector2d<float> &img,const char *fName)
{
   int w=img.w, h=img.h;
   std::ofstream file(fName,std::ios_base::binary);
   file<<"P5\n"<<w<<" "<<h<<"\n255\n";
   float *fa=new float[w*h]; img.read(fa);
   char *ca=new char[w*h];
   for (int j=0;j<h;j++)
      for (int i=0;i<w;i++)
         ca[i+j*w]=(unsigned char)(255.0*fa[i+j*w]);
   file.write(ca,w*h);
   delete[] fa; delete[] ca;
}
/****************** OpenCL basics ****************/

/* Dimensions for all 2D examples below: */
int w=1024, h=1024;


/* Show deviations of reality from timing model */
bool show_deviations=false;


GPU_STRUCT(vec2, // <- struct's name
// OpenCL and C++ declaration of fields
	float x; 
	float y;
,
// C++-only methods
	vec2() {}
	vec2(float x_,float y_) { x=x_; y=y_; }
	vec2 &operator+=(const vec2 &rhs) {x+=rhs.x; y+=rhs.y; return *this; }
	friend vec2 operator+(vec2 lhs,const vec2 &rhs) {lhs+=rhs; return lhs;}
)
GPU_STRUCT(vector2d_info, // <- struct's name
// OpenCL and C++ declaration of fields
	__global<float *> base; // base address (cl_mem pointer)
	int w; // width (i or x size)
	int h; // height (j or y size)
,
// C++-only methods (empty)
	vector2d_info() {base=0; w=h=0; }
	vector2d_info(__global<float *> base_,int w_,int h_)
		{ base=base_; w=w_; h=h_; }
)

#if BENCH_ATOMICS
GPU_ADD_STATIC_CODE(
"#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable\n"
,extension_atomic_inc_pragma);
#endif

#if BENCH_DOUBLE
GPU_ADD_STATIC_CODE(
"#pragma OPENCL EXTENSION cl_khr_fp64: enable\n"
,extension_fp64_pragma);
#endif

/******************** Basic Kernel Testing *****************/

int length=0;
enum {max_length=1<<22};
int host_vector[max_length];
float host_vectorf[max_length];
gpu_vector<int> device_vector(max_length);
gpu_vector<int> device_vector_host(max_length,(int *)gpu_buffer::alloc_host);
gpu_vector<float> device_vectorf(max_length);
#if BENCH_DOUBLE
gpu_vector<double> device_vectord(max_length);
#endif

int just_clFinish(void) {clFinish(gpu_env::static_env().clQUE); return 0;}

int allocate_cvector(void) { gpu_vector<char> arr(length); return arr.size(); }
int A_ivector(void) { gpu_vector<int> arr(length); return arr.size(); }
int allocate_dvector(void) { gpu_vector<double> arr(length); return arr.size(); }

int R_ivector(void) { device_vector.read(&host_vector[0],0,length); return 0; }
int W_ivector(void) { device_vector.write(&host_vector[0],0,length); return 0; }

int R_fvector(void) { device_vectorf.read(&host_vectorf[0],0,length); return 0; }
int W_fvector(void) { device_vectorf.write(&host_vectorf[0],0,length); return 0; }

int MR_ivector(void) { 
	int *ptr=device_vector.map(CL_MAP_READ,0,length); int ret=ptr[0]; device_vector.unmap();
	return ret; 
}
int MW_ivector(void) { 
	int *ptr=device_vector.map(CL_MAP_WRITE,0,length); ptr[0]++; device_vector.unmap();
	return 0; 
}
int MRW_ivector(void) { 
	int *ptr=device_vector.map(CL_MAP_READ+CL_MAP_WRITE,0,length); ptr[0]++; device_vector.unmap();
	return 0; 
}
int HMRW_ivector(void) { 
	int *ptr=device_vector_host.map(CL_MAP_READ+CL_MAP_WRITE,0,length); ptr[0]++; device_vector_host.unmap();
	return 0; 
}


int AWR_ivector(void) { 
	gpu_vector<int> arr(length); 
	arr.write(&host_vector[0],0,length); 
	arr.read(&host_vector[0],0,length); 
	return 0; 
}
int WR_ivector(void) { 
    device_vector.write(&host_vector[0],0,length); 
    device_vector.read(&host_vector[0],0,length); 
	return 0; 
}

GPU_FILLKERNEL(int, incInt, (), result++; )

int AWKR_ivector(void) { 
	gpu_vector<int> arr(length); 
	arr.write(&host_vector[0],0,length);
	arr=incInt();
	arr.read(&host_vector[0],0,length); 
	return 0; 
}
int WKR_ivector(void) { 
	device_vector.write(&host_vector[0],0,length);
	device_vector=incInt();
	device_vector.read(&host_vector[0],0,length);
	return 0; 
}
int K_ivector(void) { 
	device_vector=incInt();
	return 0; 
}
GPU_FILLKERNEL(int, incInt1, (int a), result+=a; )
int K1_ivector(void) { 
	device_vector=incInt1(1);
	return 0; 
}
GPU_FILLKERNEL(int, incInt2, (int a,int b), result=result*a+b; )
int K2_ivector(void) { 
	device_vector=incInt2(1,1);
	return 0; 
}
GPU_FILLKERNEL(int, incInt3, (int a,int b,__global<int *> arr), result=arr[i]*a+b; )
int K3_ivector(void) { 
	device_vector=incInt3(1,1,device_vector);
	return 0; 
}
int AK_ivector(void) { 
	gpu_vector<int> arr(length); 
	arr=incInt();
	return 0; 
}
int AK2_ivector(void) { 
	gpu_vector<int> arr1(length); 
	gpu_vector<int> arr2(length); 
	arr1=incInt();
	arr2=incInt();
	return 0; 
}
int AK3_ivector(void) { 
	gpu_vector<int> arr1(length); 
	gpu_vector<int> arr2(length); 
	gpu_vector<int> arr3(length);
	arr1=incInt();
	arr2=incInt();
	arr3=incInt();
	return 0; 
}

GPU_FILLKERNEL(float, incFloat, (), result+=1.0f; )
int K_fvector(void) { 
	device_vectorf=incFloat();
	return 0; 
}
int AK_fvector(void) { 
	gpu_vector<float> arr(length); 
	arr=incFloat();
	return 0; 
}

GPU_FILLKERNEL(float, fpoly1, (float a,float b), 
	result=b+result*a; 
)
int poly1(void) { 
	device_vectorf=fpoly1(0.1,0.3);
	return 0; 
}

#if BENCH_DOUBLE
GPU_FILLKERNEL(double, dpoly1, (double a,double b), 
	result=b+result*a; 
)
int poly1d(void) { 
	device_vectord=dpoly1(0.1,0.3);
	return 0; 
}
#endif

GPU_FILLKERNEL(float, fpoly2, (float a,float b,float c), 
	result=c+result*(b+result*a); 
)
int poly2(void) { 
	device_vectorf=fpoly2(-0.01,0.1,0.3);
	return 0; 
}
GPU_FILLKERNEL(float, fpoly3, (float a,float b,float c,float d), 
	result=d+result*(c+result*(b+result*a)); 
)
int poly3(void) { 
	device_vectorf=fpoly3(0.0000001,-0.01,0.01,0.2);
	return 0; 
}
GPU_FILLKERNEL(float, fpoly4, (float a,float b,float c,float d,float e), 
	result=e+result*(d+result*(c+result*(b+result*a))); 
)
int poly4(void) { 
	device_vectorf=fpoly4(1.0e-10,0.0000001,-0.01,0.01,0.2);
	return 0; 
}

/* "Teaser" figure in white paper */
GPU_FILLKERNEL(float,  // OpenCL code
 poly3gpu, (float a,float b,float c,float d), 
 {
   if (result>=0) {
     result=((a*result+b)*result+c)*result+d;
   } 
 }
)
void poly3(gpu_vector<float> &arr) { // C++
   arr=poly3gpu(0.0000001,-0.01,0.01,0.2);
}
int paperdemo(void) {
	poly3(device_vectorf);
	return 0;
}

double time_length(timeable_fn fn,size_t len) {
	length=len;
	device_vector.remark_size(length);
	device_vectorf.remark_size(length);
	if (length<10000) fn(); /* run once (warm up caches) */
	double ret=time_function(fn);

	/* Do a test read, to immediately get any kernel execution errors */
		int i=0;
		device_vector.read(&i,0,1);
	
	device_vector.remark_size(max_length);
	device_vectorf.remark_size(max_length);
	return ret;
}

/**
  Run this function with length between 1<<lo and 1<<hi.
*/
void model_alphabeta(timeable_fn fn,const char *what,int lo,int hi) {
	while ((1<<hi)>max_length) hi--; /* back off high to fit in max_length */
	double t[64]={0.0};
	
	t[lo]=time_length(fn,1<<lo);
	t[hi]=time_length(fn,1<<hi);
	
	double alpha=t[lo], beta=t[hi]/(1<<hi); // beta=(t[hi]-t[lo])/((1<<hi)-(1<<lo));
	
	printf("%20s: %.3f us + %.3f ns per\n",what,alpha*1.0e6,beta*1.0e9);
	fflush(stdout);
	
	if (show_deviations) { /* compare reality against model */
		for (int i=lo+1;i<hi;i++) {
			t[i]=time_length(fn,1<<i);
			int length=1<<i;
			double model=alpha+beta*length;
			double diff=t[i]-model;
			if (fabs(diff)/model>0.30) { /* over a 30% deviation--print */
				printf("	                but len=%d -> t=%.3f us (%+.0f%%)\n",
					length,t[i]*1.0e6,100.0*diff/model);
			}
		}
	}
}

GPU_FILLKERNEL(float,  wgTest, (float k), result+=k   );
int test_workgroup(void) {
	device_vectorf=wgTest(0.001);
	return 0;
}

double time_workgroup(int width) {
	wgTest.override_local[0]=width;
	char buf[100]; sprintf(buf,"workgroup<%d>",width);
	//model_alphabeta(test_workgroup,buf,0,24); // alpha basically unchanged
	size_t len=max_length;
	double nsper=time_length(test_workgroup,len)/len*1.0e9;
	return nsper;
}

/* Measure the effect of workgroup size on write bandwidth */
void measure_workgroup() {
	int sz=wgTest.get_workgroupsize(100000);
	for (int pass=0;pass<=1;pass++) {
		printf(pass?"ns/float":"workgroup");
		printf(" &");
		for (int width=1;width<=sz;width*=2) {
			if (pass==0) printf("%d	&",width);
			else printf("%4.3f	&",time_workgroup(width));
		}
		printf("\\\\\\hline\n");
	}
	FILE *f=fopen("bench_workgroup.txt","w");
	for (int width=1;width<=sz;width++) {
		double ns=time_workgroup(width);
		fprintf(f,"%d	%4.3f\n",width,ns); fflush(f);
		printf("%d	%4.3f\n",width,ns); fflush(stdout);
	}
	fclose(f);
}



// Print performance for wxh pixels of work done in this function.
//    bytesper and flopsper are memory and arithmetic used per pixel.
double perf_wxh(timeable_fn fn,const char *what,int bytesper,int flopsper) {
	length=w*h;
	
	double nsper=time_function(fn)/(w*h)*1.0e9;
	printf("%20s: %.3f ns/pixel, %.3f GB/s, %.3f GF/s\n",
		what,nsper,bytesper/nsper,flopsper/nsper);
	fflush(stdout);
	return nsper;
}


/********************** Overheads from Access Patterns ****************/

#if BENCH_ATOMICS
/* Percent non-contended atomic accesses */
int atomic_percent=100;
int atomic_mask=0;
int atomic_xor=0;

GPU_KERNEL( atom_inc_iK, (__global<int *> arr,int percent,int mask,int bxor), atom_inc(&arr[mask&(bxor^(i*percent/100))]); )
int atom_inc_i(void) { 
	atom_inc_iK(device_vector,atomic_percent,~atomic_mask,atomic_xor).run(w*h); 
	return 0; 
}

int atomic_add=1;
GPU_KERNEL( atom_add_iK, (__global<int *> arr,int percent,int val), atom_add(&arr[i*percent/100],val); )
int atom_add_i(void) { 
	atom_add_iK(device_vector,atomic_percent,atomic_add).run(w*h); 
	return 0; 
}

/* OpenCL doesn't have atomic float ops yet, so fake 'em with compare-and-swap 
  Sadly, this crashes CUDA
  http://www.khronos.org/registry/cl/sdk/1.1/docs/man/xhtml/atomic_cmpxchg.html
  http://www.khronos.org/registry/cl/sdk/1.0/docs/man/xhtml/as_typen.html
*/
float atomic_addf=1.12f;
GPU_KERNEL( atom_add_fK, (__global<float *> arr,int percent,float val), 
{
	__global float *fp=&arr[i*percent/100];
	__global int *ip=fp;
	int old_i=0;  // value to operate on
	int last_i=*ip; // value last read from memory
	do {
		old_i=last_i; 
		float old=as_float(old_i);
		float nu=old+val;
		last_i=atom_cmpxchg(ip,old_i,as_int(nu));
	} while (old_i!=last_i);
})
int atom_add_f(void) { 
	atom_add_fK(device_vectorf,atomic_percent,atomic_addf).run(w*h); 
	return 0; 
}

void atomic_sweep(timeable_fn fn,const char *what) {
	length=w*h;
	
	printf("%15s: ",what);
	int pct[]={101,100,99,90,0};
	int pi=0;
	do {
		atomic_percent=pct[pi];
		double nsper=time_function(fn)/(w*h)*1.0e9;
		printf("%.3f ns (%d%%), ",nsper, pct[pi]); fflush(stdout);
	} while (pct[pi++]!=0);
	printf("\n");
	fflush(stdout);
}


#endif







/* Non-EPGPU style code (to test overhead) */
GPU_ADD_STATIC_CODE(
"__kernel void write1DplainK(__global int *arr) {arr[get_global_id(0)]=1;}\n",
write1DplainK
)
int write1Dplain(void) { 
	static cl_program theProg=gpu_env::static_env().all_compiled();
	static cl_int err=0;
	static cl_kernel theKrnl=clCreateKernel(theProg, "write1DplainK", &err);
	cl_mem m=device_vector;
	clSetKernelArg(theKrnl, 0, sizeof(m), &m);
	size_t localsz = 256;
	size_t globalsz = (length+localsz-1)/localsz*localsz;
	clEnqueueNDRangeKernel(gpu_env::static_env().clQUE, theKrnl, 1, NULL, 
                            	 &globalsz, &localsz, 0, 
                            	 NULL, NULL);
	return 0; 
}

/* EPGPU version of above */
GPU_KERNEL(
write1DK, (__global<int *> arr), 
	arr[i]=1;
)
int write1D(void) { 
	write1DK(device_vector).run(w*h);
	return 0; 
}


/* EPGPU FILL version of above */
GPU_FILLKERNEL(int, write1DfillK, (), {result=1;} )
int write1Dfill(void) { 
	device_vector.remark_size(length);
	device_vector=write1DfillK();
	device_vector.remark_size(max_length);
	return 0; 
}




GPU_KERNEL(
write2DrowsinterK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[i+w*j]=interleave2D(i,j);
)
int write2Drowsinter(void) { 
	write2DrowsinterK(w,h,device_vector).run(w,h);
	return 0; 
}
GPU_KERNEL(
write2DrowsK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[i+w*j]=1;
)
int write2Drows(void) { 
	write2DrowsK(w,h,device_vector).run(w,h);
	return 0; 
}
GPU_KERNEL(
write2DcolsK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[j+w*i]=1;
)
int write2Dcols(void) { 
	write2DcolsK(w,h,device_vector).run(w,h);
	return 0; 
}

GPU_KERNEL(
write2DinterK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[interleave2D(i,j)]=1;
)
int write2Dinter(void) { 
	write2DinterK(w,h,device_vector).run(w,h);
	return 0; 
}
GPU_KERNEL(
write2DinterjiK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[interleave2D(j,i)]=1;
)
int write2Dinterji(void) { 
	write2DinterjiK(w,h,device_vector).run(w,h);
	return 0; 
}


GPU_KERNEL(
rmw1DK, (__global<int *> arr), 
	arr[i]+=1;
)
int rmw1D(void) { 
	rmw1DK(device_vector).run(w*h);
	return 0; 
}

GPU_KERNEL(
rmw2DrowsK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[i+w*j]+=1;
)
int rmw2Drows(void) { 
	rmw2DrowsK(w,h,device_vector).run(w,h);
	return 0; 
}
GPU_KERNEL(
rmw2DcolsK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[j+w*i]+=1;
)
int rmw2Dcols(void) { 
	rmw2DcolsK(w,h,device_vector).run(w,h);
	return 0; 
}

GPU_KERNEL(
rmw2DinterK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[interleave2D(i,j)]+=1;
)
int rmw2Dinter(void) { 
	rmw2DinterK(w,h,device_vector).run(w,h);
	return 0; 
}

GPU_KERNEL(
rmw2DinterjiK, (int w,int h,__global<int *> arr), 
	int j=get_global_id(1);
	arr[interleave2D(j,i)]+=1;
)
int rmw2Dinterji(void) { 
	rmw2DinterjiK(w,h,device_vector).run(w,h);
	return 0; 
}

gpu_vector2d<float> transpose_src(h,w),transpose_dst(w,h);

GPU_KERNEL(
transpose2DinterK, (__global<float *> dst,__global<float *> src), 
	int j=get_global_id(1);
	dst[interleave2D(j,i)]=src[interleave2D(i,j)];
)
int transpose2Dinter(void) { 
	transpose2DinterK(transpose_dst,transpose_src).run(w,h);
	return 0; 
}

/********************** Transpose example ****************/

GPU_FILLKERNEL_2D(float,
 naive_transposeGPU,(__global<float *> src),
   result=src[j+i*h];
);

//gpu_vector2d<float> transpose_src(h,w),transpose_dst(w,h);
int naive_transpose(void) {
	transpose_dst=naive_transposeGPU(transpose_src);
	return 0;
}

/**
  Transpose 2D src h x w into 2D dst w x h.
      dst[i+j*w] = src[j+i*h];
  To avoid global memory scatter, does transposes to local memory.
*/
GPU_KERNEL_2D(local_transposeGPU,
 (__global<float *> dst,__global<float *> src,
  int w,int h),
   enum {N=16}; // size of local work groups
   int2 G=N*(int2)(get_group_id(0),get_group_id(1));
   int2 L=  (int2)(get_local_id(0),get_local_id(1));
   __local float loc[N*N]; // transpose here
   
   int2 S=G+(int2)(L.y,L.x); // Read backwards
   if (S.x<w && S.y<h) 
     loc[L.y+N*L.x]=src[S.y+S.x*h];
   
   barrier(CLK_LOCAL_MEM_FENCE);
   
   int2 D=G+L; // Write forwards
   if (D.x<w && D.y<h) 
     dst[D.x+D.y*w]=loc[L.x+N*L.y];
);

int local_transpose(void) {
	local_transposeGPU.override_local[0]=16;
	local_transposeGPU.override_local[1]=16;
	local_transposeGPU(transpose_dst,transpose_src,w,h).run(w,h);
	return 0;
}

/********************** Stencil example ****************/

/* Set up initial conditions */
GPU_FILLKERNEL_2D(float,
stencil_initial,(float sz,float cx,float cy,float r),
   float x=i*sz-cx; float y=j*sz-cy;
   if (fmod(sqrt(x*x+y*y),2*r)<=r) {
      result=1.0;
   } else {
      result=0.00001; //<- avoid denormals
   }
)

/* Do neighborhood averages */
/* all 2D indexing */
GPU_FILLKERNEL_2D(float,stencil_sweep_2D,(__global<float *> src),
   if (i>0 && i<w-1 && j>0 && j<h-1) { // Interior
      result = 0.25*(src[i-1+w*j]+src[i+1+w*j]+src[i+w*(j-1)]+src[i+w*(j+1)]);
   } else { // Boundary--just copy existing value 
      result = src[i+w*j];
   }
)

// 2D to 1D indexing
GPU_FILLKERNEL_2D(float,stencil_sweep_1D,(__global<float *> src),
   int n=i+w*j; // 2D to 1D indexing
   if (i>0 && i<w-1 && j>0 && j<h-1) {
      result = (src[n-1]+src[n+1]+
	            src[n-w]+src[n+w])*0.25; // <- divide is slower on ATI: /4.0;
   } else { // Boundary--copy old value 
      result = src[n];
   }
)

/* Morton-order indexing */
GPU_KERNEL_2D(stencil_morton,(__global<float *> src,__global<float *> dest,int w,int h),
   if (i>0 && i<w-1 && j>0 && j<h-1) { // Interior
   	  dest[interleave2D(i,j)]=0.25*(
	  	src[interleave2D(i-1,j)]+
	  	src[interleave2D(i+1,j)]+
	  	src[interleave2D(i,j-1)]+
	  	src[interleave2D(i,j+1)]
	  );
   } // Boundary--just leave existing value 
)

/* Steamroller local storage: wide and flat chunks
  This still a *big* win, even on Fermi, unlike the obvious ones below.
256x1: 54GB/s (similar to naive 16x16)
256x2: 70GB/s
256x4: 78GB/s
256x8: 78GB/s -> 4-8 rows/workgroup is best
256x16: 66GB/s

128x8: 68GB/s 
256x8: 78GB/s -> 256-wide workgroups best
512x8: 72GB/s 
1024x8: 48GB/s

128x16: 62GB/s

  Derived from this answer:
  http://stackoverflow.com/questions/5794617/cuda-shared-memory-over-a-large-ish-2d-array
*/
enum {stencil_roller_iwk=256}; // workgroup width (height is always 1)
enum {stencil_roller_rowsper=8}; // MUST match nrows in OpenCL...

// 4-neighbor specialized roller
GPU_KERNEL_2D(stencil_roller,(__global<float *> src,__global<float *> dest,int w,int h),
	enum {iwk=256}; // workgroup size, in i
	enum {sw=iwk}; // dimensions of local workgroup
	enum {nrows=8}; // number of rows for each workgroup to perform
	i-=(i/iwk)*2; // close up gaps due to edge pixels
	j*=nrows; /* only one execution for every j rows */
	//if (j+nrows>h) return; /* we fell off the end */
	
	__local float L[sw]; // local neighbor-read buffer
	int x=get_local_id(0); int gidx=(i)+w*(j);
	int lastidx=gidx-w;
	if (j-1<0) lastidx=gidx; // there is no previous row, so use this row
	float prev=src[lastidx]; /* fetch previous row */
	float cur =src[gidx]; /* fetch current row */
	for (int row=0;row<nrows;row++) {
		// Read next row
		float next=0.0;
		if (j+row+1<h) next=src[gidx+(row+1)*w];
		
		// Write current row to shared memory (for neighbor accesses below)
		L[x]=cur;
		barrier(CLK_LOCAL_MEM_FENCE); // finish all writes before doing reads
		
		// Compute outputs (for non-edge pixels)
		if (x>0 && x<iwk-1)
			dest[gidx+(row)*w]= 0.25*(
				L[x-1]+L[x+1]+next+prev
			);
		//barrier(CLK_LOCAL_MEM_FENCE); // finish all reads before doing next writes (needed?)
		
		// roll over registers
		prev=cur;
		cur=next;
	}
)

/*
 2 timestep roller
 256x8: 56GB/s*2
 256x16: 62GB/s*2
 256x24: 63GB/s*2
 256x32: 65GB/s*2
 256x48: 60GB/s*2
 256x64: 62GB/s*2
 
 512x16: 56GB/s*2
 128x32: 62GB/s*2
*/
enum {stencil_roller2_iwk=256}; // workgroup width (height is always 1)
enum {stencil_roller2_rowsper=32}; // MUST match nrows in OpenCL...
GPU_KERNEL_2D(stencil_roller2,(__global<float *> src,__global<float *> dest,int w,int h),
	enum {iwk=256}; // workgroup size, in i
	enum {sw=iwk}; // dimensions of local workgroup
	enum {nrows=32}; // number of rows for each workgroup to perform
	i-=(i/iwk)*4; // close up gaps due to edge pixels
	j*=nrows; /* only one execution for every j rows */
	//if (j+nrows>h) return; /* we fell off the end */
	
	__local float L[2][1+sw+1]; // local neighbor-read buffer
	int x=get_local_id(0); int gidx=(i)+w*(j);
	int j3=gidx-3*w;
	int j2=gidx-2*w;
	if (j-3<0) j3=j2; // clamp addressing to be in-bounds
	if (j-2<0) j3=j2=gidx; 
	
	float r1t0=src[j3]; /* gj-1, timestep 0 */
	float r0t0=src[j2]; /* gj-0, timestep 0 */
	float r1t1=0; /* row-1, timestep 1 */
	float r2t1=0; /* row-2, timestep 1 */

/*
Time is vertical, space (row j) is horizontal:
 t2       dest     
       /    |   \        
 t1  r2t1 r1t1 r0t1                 
            /   |   \ 
 t0       r1t0 r0t0 rNt0(src)
                gj        
*/
	for (int gj=j-2;gj<j+nrows;gj++) {
		float r0t1=0; /* row 0, timestep 1 */
	
		// Read next row
		float rNt0=src[i+  min(max(gj+1,0),h-1)  *w];
		
		// Write row0 to shared memory (for time-1's neighbor reads)
		L[0][1+x]=r0t0;
		
		// Write row1, time 1 to shared memory (for time-2's neighbors)
		L[1][1+x]=r1t1;
		
		barrier(CLK_LOCAL_MEM_FENCE); // finish all writes before doing reads
		
		// Compute outputs (for non-edge pixels, to avoid bad side reads)
		//if (x>0 && x<iwk-1) { // <- slowdown, replaced with 1+ on L[x]
			// time-1 outputs to registers
			r0t1=0.25*(
				L[0][x]+L[0][x+2]+rNt0+r1t0
			);
			
			// do time-2 outputs to memory
			float r1t2=0.25*(
				L[1][x]+L[1][x+2]+r0t1+r2t1
			);
			if (x>1 && x<iwk-2 // edges don't have all X data
				&& gj>j-1  // r2t1 isn't ready yet
				&& gj-1>=0 && gj-1<h // out of bounds anyway
			) { // do time-2 outputs to memory
				dest[i+(gj-1)*w]=r1t2;
			}
		//}
		//barrier(CLK_LOCAL_MEM_FENCE); // finish all reads before doing next writes (needed?)
		
		// roll down register blocks
		r1t0=r0t0; r0t0=rNt0;
		r2t1=r1t1; r1t1=r0t1;
	}
)


// 8-point compatible roller
GPU_KERNEL_2D(stencil_roller8,(__global<float *> src,__global<float *> dest,int w,int h),
	int si=get_local_id(0); // local index
	enum {iwk=256}; // workgroup size, in i
	enum {sw=iwk}; // dimensions of local workgroup
	enum {nrows=8}; // number of rows for each workgroup to perform
	i-=(i/iwk)*2; // close up gaps due to edge pixels
	j*=nrows; /* only one execution for every j rows */
	//if (j+nrows>h) return; /* we fell off the end */
	
	__local float L[3][sw]; // rolling local buffer
	int x=si; int gidx=(i)+w*(j);
	int lastidx=gidx-w;
	if (j-1<0) lastidx=gidx; // there is no previous row, so use this row
	L[0][x]=src[lastidx]; /* fetch previous row */
	L[1][x]=src[gidx]; /* fetch current row */
	unsigned int prev=0; // L index of previous row
	unsigned int cur=1;  // L index of current row
	unsigned int next=2; // L index of next row
	
	for (int row=0;row<nrows;row++) {
		// Read next row
		if ((j+row+1)<h)
			L[next][x]=src[gidx+(row+1)*w];
		barrier(CLK_LOCAL_MEM_FENCE); // finish all writes
		
		// Compute outputs (for non-edge pixels)
		if (x>0 && x<iwk-1)
			dest[gidx+(row)*w]= 0.25*(
				L[cur][x-1]+L[cur][x+1]+L[next][x]+L[prev][x]
			);
		barrier(CLK_LOCAL_MEM_FENCE); // finish all reads before doing next writes
		
		if (false) { /* Circular buffer (slightly faster, but more complex) */
			prev=(prev+1)%3; 
			cur =(cur +1)%3; 
			next=(next+1)%3; 
		} else { /* shuffle data up to next row (slightly slower, but simpler) */
			L[prev][x]=L[cur ][x];
			L[cur ][x]=L[next][x];
		}
	}
)


// local: everybody stores
GPU_FILLKERNEL_2D(float,stencil_local,(__global<float *> src),
	int si=get_local_id(0)+1; int sj=get_local_id(1)+1; // local
	enum {sw=16+2}; 
	enum {sh=16+2}; // dimensions of local workgroup, plus 1 on each side
	
	__local float L[sw*sh]; // local(si,sj) stores src(i,j)
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	L[sidx]=src[gidx];
	
	// Fetch boundaries
	if (si==1    && (i-1>=0)) L[sidx-1]=src[gidx-1];
	if (si==sw-2 && (i+1<w))  L[sidx+1]=src[gidx+1];
	if (sj==1    && (j-1>=0)) L[sidx-sw]=src[gidx-w];
	if (sj==sh-2 && (j+1<h))  L[sidx+sw]=src[gidx+w];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local writes
	
	if (i>0 && i<w-1 && j>0 && j<h-1) { // interior
		result=0.25*(
			L[sidx-1]+L[sidx-sw]+L[sidx+1]+L[sidx+sw]
		);
	} else { // Boundary--just copy old value 
		result=L[sidx];
	}
)

// 14x14 output local: everybody loads, everybody stores *except* boundary
GPU_KERNEL_2D(stencil_local14,(__global<float *> src,__global<float *> dest,int w,int h),
	//int i=get_global_id(0); int j=get_global_id(1); // global
	enum {sw=16}; 
	enum {sh=16}; // dimensions of local workgroup
	i-=(i/sw)*2; // close up spaces between valid output regions
	j-=(j/sh)*2;
	
	int si=get_local_id(0); int sj=get_local_id(1); // local
	
	__local float L[sw*sh]; // local(si,sj) stores src(i,j)
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	if (i<w && j<h)
		L[sidx]=src[gidx];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local writes
	
	if (si<=0 || si>=sw-1 || sj<=0 || sj>=sh-1) return; // only there for boundary load
	// else we are interior
	//if (i<w-1 && j<h-1)
		dest[gidx]=0.25*(
			L[sidx-1]+L[sidx-sw]+
			L[sidx+1]+L[sidx+sw]
		);
)

// 12x12 output two-timestep local.  FAST AND LOOSE on boundaries--beware!
GPU_KERNEL_2D(stencil_local12,(__global<float *> src,__global<float *> dest,int w,int h),
	//int i=get_global_id(0); int j=get_global_id(1); // global
	enum {iwk=16}; 
	enum {jwk=16}; // dimensions of local workgroup
	i-=(i/iwk)*4; // close up spaces between valid output regions
	j-=(j/jwk)*4;
	
	int si=1+get_local_id(0); int sj=1+get_local_id(1); // local
	
	enum {sw=iwk+2}; 
	enum {sh=jwk+2};
	__local float L[2*sw*sh]; // local(si,sj) stores src(i,j)
	int scratchboost=sw*sh; // distance to other local buffer
	
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	if (i<w && j<h)
		L[sidx]=src[gidx];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local loads
	
	//if (si>1 && si<sw-2 && sj>1 && sj<sh-2) 
	{ // interior
		L[scratchboost+sidx]=0.25*(
			L[sidx-1]+L[sidx-sw]+
			L[sidx+1]+L[sidx+sw]
		);
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local computes
	
	if (si>2 && si<sw-3 && sj>2 && sj<sh-3) 
	{ // interior + 1
		sidx+=scratchboost;
		dest[gidx]=0.25*(
			L[sidx-1]+L[sidx-sw]+
			L[sidx+1]+L[sidx+sw]
		);
	}
)

// 10x10 output three-timestep local.  FAST AND LOOSE on boundaries--beware!
GPU_KERNEL_2D(stencil_local10,(__global<float *> src,__global<float *> dest,int w,int h),
	//int i=get_global_id(0); int j=get_global_id(1); // global
	enum {iwk=16}; 
	enum {jwk=16}; // dimensions of local workgroup
	i-=(i/iwk)*6; // close up spaces between valid output regions
	j-=(j/jwk)*6;
	
	int si=1+get_local_id(0); int sj=1+get_local_id(1); // local
	
	enum {sw=iwk+2}; 
	enum {sh=jwk+2};
	enum {off=5}; // to prevent bank conflicts?
	__local float L[off+2*sw*sh]; // local(si,sj) stores src(i,j)
	int scratchboost=off+sw*sh; // distance to Jacobi farside of local buffer
	
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	int s2=sidx+scratchboost;
	if (i<w && j<h)
		L[sidx]=src[gidx];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local loads
	
	//if (si>0 && si<sw-1 && sj>0 && sj<sh-1) 
	{ // interior
		L[s2]=0.25*(
			L[sidx-1]+L[sidx-sw]+
			L[sidx+1]+L[sidx+sw]
		);
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local loads
	
	//if (si>1 && si<sw-2 && sj>1 && sj<sh-2) 
	{ // interior + 1
		L[sidx]=0.25*(
			L[s2-1]+L[s2-sw]+
			L[s2+1]+L[s2+sw]
		);
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local computes
	
	if (si>2 && si<sw-3 && sj>2 && sj<sh-3) 
	{ // interior + 2
		dest[gidx]=0.25*(
			L[sidx-1]+L[sidx-sw]+
			L[sidx+1]+L[sidx+sw]
		);
	}
)
gpu_vector2d<float> stencil_src(w,h); // gotta be global so stencil_sweep can be tested
gpu_vector2d<float> stencil_dst(w,h);

int sweep_type;
int stencil_sweep(void) {
   switch (sweep_type) {
   case 0: stencil_dst=stencil_sweep_2D(stencil_src); break;
   case 1: stencil_dst=stencil_sweep_1D(stencil_src); break;
   case 2: stencil_roller(stencil_src,stencil_dst,
			stencil_dst.w,stencil_dst.h)
		 	.run(stencil_dst.w,stencil_dst.h/stencil_roller_rowsper);
		   break;
   case 3: stencil_roller2(stencil_src,stencil_dst,
			stencil_dst.w,stencil_dst.h)
		 	.run(stencil_dst.w,stencil_dst.h/stencil_roller2_rowsper);
		   break;
   case 4: stencil_morton(stencil_src,stencil_dst,
			stencil_dst.w,stencil_dst.h)
		 	.run(stencil_dst.w,stencil_dst.h);
		   break;
   case 5: stencil_dst=stencil_local(stencil_src); break;
   case 6:
   		stencil_local14(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w/14*16,stencil_dst.h/14*16); 
		break;
   case 7:
   		stencil_local12(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w/12*16,stencil_dst.h/12*16); 
		break;
   case 8:
   		stencil_local10(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w/10*16,stencil_dst.h/10*16); 
		break;
   }
   std::swap(stencil_dst,stencil_src); /* <- ping-pongs the buffers */
   return 0;
}

int stencil_run(void) {
	stencil_roller.override_local[0]=stencil_roller_iwk;
	stencil_roller.override_local[1]=1;
	stencil_roller2.override_local[0]=stencil_roller2_iwk;
	stencil_roller2.override_local[1]=1;

	stencil_local.override_local[0]=16;
	stencil_local.override_local[1]=16;
	stencil_local14.override_local[0]=16;
	stencil_local14.override_local[1]=16;
	stencil_local12.override_local[0]=16;
	stencil_local12.override_local[1]=16;
	stencil_local10.override_local[0]=16;
	stencil_local10.override_local[1]=16;

	const char *sweep_name[]={"naive2D","naive1D",
		"roller","roller2",
		"morton",
		"local16","local14","local12","local10",0};
	for (sweep_type=0;sweep_name[sweep_type];sweep_type++) {
		stencil_src=stencil_initial(0.01,6.0,2.4,3.0);
		stencil_dst=stencil_initial(0.01,6.0,2.4,3.0);
		for (int time=0;time<1000;time++) stencil_sweep();
		
		if (false) {
			// Test out fancy transpose while we're at it...
			gpu_vector2d<float> stencil_dst(h,w);
			local_transposeGPU.override_local[0]=16;
			local_transposeGPU.override_local[1]=16;
			local_transposeGPU(stencil_dst,stencil_src,w,h).run(w,h);
		}

		// Write data to output file
		char buf[100];
		sprintf(buf,"stencil_%s.ppm",sweep_name[sweep_type]);
		dump_image(stencil_dst,buf);
		
		// Do a performance run
		perf_wxh(stencil_sweep,buf,sizeof(float)*5,4);
	}
	return 0;
}

/******* 8-point all neighbors 2D stencil */
/* Do neighborhood averages */
/* all 2D indexing */
GPU_FILLKERNEL_2D(float,stencil8_sweep_2D,(__global<float *> src),
   if (i>0 && i<w-1 && j>0 && j<h-1) { // Interior
      result = 0.125*(
	  	src[i-1+w*j]+src[i+1+w*j]+src[i+w*(j-1)]+src[i+w*(j+1)]+
	  	src[i-1+w*(j+1)]+src[i+1+w*(j-1)]+src[i-1+w*(j-1)]+src[i+1+w*(j+1)]
	);
   } else { // Boundary--just copy existing value 
      result = src[i+w*j];
   }
)

// 2D to 1D indexing
GPU_FILLKERNEL_2D(float,stencil8_sweep_1D,(__global<float *> src),
   int n=i+w*j; // 2D to 1D indexing
   if (i>0 && i<w-1 && j>0 && j<h-1) {
      result = (src[n-1]+src[n+1]+
	            src[n-w]+src[n+w]+
				src[n+1-w]+src[n+1+w]+
	            src[n-1-w]+src[n-1+w])*0.125; // <- divide is slower on ATI: /4.0;
   } else { // Boundary--copy old value 
      result = src[n];
   }
)

/* Morton-order indexing */
GPU_KERNEL_2D(stencil8_morton,(__global<float *> src,__global<float *> dest,int w,int h),
   if (i>0 && i<w-1 && j>0 && j<h-1) { // Interior
   	  dest[interleave2D(i,j)]=0.125*(
	  	src[interleave2D(i-1,j)]+
	  	src[interleave2D(i+1,j)]+
	  	src[interleave2D(i,j-1)]+
	  	src[interleave2D(i,j+1)]+
	  	src[interleave2D(i-1,j+1)]+
	  	src[interleave2D(i+1,j-1)]+
	  	src[interleave2D(i-1,j-1)]+
	  	src[interleave2D(i+1,j+1)]
	  );
   } // Boundary--just leave existing value 
)

// local: everybody stores
GPU_FILLKERNEL_2D(float,stencil8_local,(__global<float *> src),
	int si=get_local_id(0)+1; int sj=get_local_id(1)+1; // local
	enum {sw=16+2}; 
	enum {sh=16+2}; // dimensions of local workgroup, plus 1 on each side
	
	__local float L[sw*sh]; // local(si,sj) stores src(i,j)
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	L[sidx]=src[gidx];
	
	// Fetch boundaries
	if (si==1    && (i-1>=0)) L[sidx-1]=src[gidx-1];
	if (si==sw-2 && (i+1<w))  L[sidx+1]=src[gidx+1];
	if (sj==1    && (j-1>=0)) L[sidx-sw]=src[gidx-w];
	if (sj==sh-2 && (j+1<h))  L[sidx+sw]=src[gidx+w];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local writes
	
	if (i>0 && i<w-1 && j>0 && j<h-1) { // interior
		result=0.125*(
			L[sidx-1]+L[sidx-sw]+L[sidx+1]+L[sidx+sw]+
			L[sidx-1-sw]+L[sidx+1-sw]+L[sidx+1+sw]+L[sidx-1+sw]
		);
	} else { // Boundary--just copy old value 
		result=L[sidx];
	}
)

// 14x14 output local: everybody loads, everybody stores *except* boundary
GPU_KERNEL_2D(stencil8_local14,(__global<float *> src,__global<float *> dest,int w,int h),
	//int i=get_global_id(0); int j=get_global_id(1); // global
	enum {sw=16}; 
	enum {sh=16}; // dimensions of local workgroup
	i-=(i/sw)*2; // close up spaces between valid output regions
	j-=(j/sh)*2;
	
	int si=get_local_id(0); int sj=get_local_id(1); // local
	
	__local float L[sw*sh]; // local(si,sj) stores src(i,j)
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	if (i<w && j<h)
		L[sidx]=src[gidx];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local writes
	
	if (si<=0 || si>=sw-1 || sj<=0 || sj>=sh-1) return; // only there for boundary load
	// else we are interior
	//if (i<w-1 && j<h-1)
		dest[gidx]=0.125*(
			L[sidx-1]+L[sidx-sw]+L[sidx+1]+L[sidx+sw]+
			L[sidx-1-sw]+L[sidx+1-sw]+L[sidx+1+sw]+L[sidx-1+sw]
		);
)

// 12x12 output two-timestep local.  FAST AND LOOSE on boundaries--beware!
GPU_KERNEL_2D(stencil8_local12,(__global<float *> src,__global<float *> dest,int w,int h),
	//int i=get_global_id(0); int j=get_global_id(1); // global
	enum {iwk=16}; 
	enum {jwk=16}; // dimensions of local workgroup
	i-=(i/iwk)*4; // close up spaces between valid output regions
	j-=(j/jwk)*4;
	
	int si=1+get_local_id(0); int sj=1+get_local_id(1); // local
	
	enum {sw=iwk+2}; 
	enum {sh=jwk+2};
	__local float L[2*sw*sh]; // local(si,sj) stores src(i,j)
	int scratchboost=sw*sh; // distance to other local buffer
	
	int sidx=si+sj*sw; int gidx=(i)+w*(j);
	if (i<w && j<h)
		L[sidx]=src[gidx];
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local loads
	
	//if (si>1 && si<sw-2 && sj>1 && sj<sh-2) 
	{ // interior
		L[scratchboost+sidx]=0.125*(
			L[sidx-1]+L[sidx-sw]+L[sidx+1]+L[sidx+sw]+
			L[sidx-1-sw]+L[sidx+1-sw]+L[sidx+1+sw]+L[sidx-1+sw]
		);
	}
	
	barrier(CLK_LOCAL_MEM_FENCE); // finish local computes
	
	if (si>2 && si<sw-3 && sj>2 && sj<sh-3) 
	{ // interior + 1
		sidx+=scratchboost;
		dest[gidx]=0.125*(
			L[sidx-1]+L[sidx-sw]+L[sidx+1]+L[sidx+sw]+
			L[sidx-1-sw]+L[sidx+1-sw]+L[sidx+1+sw]+L[sidx-1+sw]
		);
	}
)

int stencil8_sweep(void) {
   switch (sweep_type) {
   case 0: stencil_dst=stencil8_sweep_2D(stencil_src); break;
   case 1: stencil_dst=stencil8_sweep_1D(stencil_src); break;
   case 2: stencil8_morton(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w,stencil_dst.h);
		   break;
   case 3: stencil_dst=stencil8_local(stencil_src); break;
   case 4:
   		stencil8_local14(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w/14*16,stencil_dst.h/14*16); 
		break;
   case 5:
   		stencil8_local12(stencil_src,stencil_dst,
			stencil_src.w,stencil_src.h)
		 	.run(stencil_dst.w/12*16,stencil_dst.h/12*16); 
		break;
   }
   std::swap(stencil_dst,stencil_src); /* <- ping-pongs the buffers */
   return 0;
}

int stencil8_run(void) {

	stencil8_local.override_local[0]=16;
	stencil8_local.override_local[1]=16;
	stencil8_local14.override_local[0]=16;
	stencil8_local14.override_local[1]=16;
	stencil8_local12.override_local[0]=16;
	stencil8_local12.override_local[1]=16;

	const char *sweep_name[]={"naive2D","naive1D","morton",
		"local16","local14","local12",0};
	for (sweep_type=0;sweep_name[sweep_type];sweep_type++) {
		stencil_src=stencil_initial(0.01,6.0,2.4,3.0);
		stencil_dst=stencil_initial(0.01,6.0,2.4,3.0);
		for (int time=0;time<1000;time++) stencil8_sweep();
		
		char buf[100];
		sprintf(buf,"stencil8_%s.ppm",sweep_name[sweep_type]);
		perf_wxh(stencil_sweep,buf,sizeof(float)*5,4);
		dump_image(stencil_dst,buf);
	}
	return 0;
}


/******* crappy image-based version.  Why is write_imagef so slow? */
/* Set up initial conditions */
GPU_FILLKERNEL_IMAGE2D(float,istencil_initial,(),
	float x=i*0.01; float y=j*0.01;
	if (sqrt(x*x+y*y)<5.0)
		result=1.0;
	else
		result=0.0;
)

/* Do neighborhood averages */
GPU_FILLKERNEL_IMAGE2D(float,istencil_sweeper,(__read_only image2d_t src),
	const sampler_t sampler=CLK_NORMALIZED_COORDS_FALSE|CLK_ADDRESS_CLAMP|CLK_FILTER_NEAREST;
	if (i>0 && i<w-1 && j>0 && j<h-1) 
		result = 0.25*(
			read_imagef(src,sampler,(int2)(i+1,j)).x+
			read_imagef(src,sampler,(int2)(i-1,j)).x+
			read_imagef(src,sampler,(int2)(i,j+1)).x+
			read_imagef(src,sampler,(int2)(i,j-1)).x
		/* // 8-point stencil: basically exact same speed!
			read_imagef(src,sampler,(int2)(i+1,j+1)).x+
			read_imagef(src,sampler,(int2)(i-1,j+1)).x+
			read_imagef(src,sampler,(int2)(i+1,j-1)).x+
			read_imagef(src,sampler,(int2)(i-1,j-1)).x
		*/
		);
)

gpu_image2d<float> istencil_src(w,h),istencil_dst(w,h);

int istencil_sweep(void) {
	istencil_dst=istencil_sweeper(istencil_src);
	std::swap(istencil_dst,istencil_src); // ping-pong the buffers
	return 0;
}

int istencil_run(void) {
	istencil_src=istencil_initial();
	istencil_dst=istencil_initial();
	for (int i=0;i<50;i++) istencil_sweep();
	std::ofstream of("stencil_image.ppm",std::ios_base::binary);
	of<<"P5\n"<<w<<" "<<h<<"\n255\n";
	float *fa=new float[w*h];
	char *ca=new char[w*h];
	istencil_src.read(fa);
	for (int j=0;j<h;j++)
		for (int i=0;i<w;i++)
			ca[i+j*w]=(unsigned char)(255.0*fa[i+j*w]);
	of.write(ca,w*h);
	delete[] fa;
	delete[] ca;
	return 0;
}

/********************** Mandelbrot example ****************/

GPU_FILLKERNEL_2D(unsigned char,mandel_render,(float sz,float xoff,float yoff),
   /* Initial conditions */
   float2 c=(float2)(i*sz+xoff,(h-1-j)*sz+yoff);
   float2 z=c;

   /* Run the mandelbrot iteration */
   int count;
   enum {max_count=250};
   for (count=0;count<max_count;count++) {
       if ((z.x*z.x+z.y*z.y)>4.0f) break;
       z=(float2)(
	      z.x*z.x-z.y*z.y + c.x,
		  2.0f*z.x*z.y + c.y
	   );
   }

   /* Figure out the output pixel color */
   result=count;
)

gpu_vector2d<unsigned char> mandel(w,h);
int mandel_pretty=1; /* if 0, render an ugly flop-intensive area; if 1, pretty area */

int mandel_renderer(void) {
   mandel=mandel_render(0.00001,mandel_pretty*0.317,0.414);
   return 0;
}

double mandelbrot(void) {
   mandel_renderer();
   
   // Write data to output file
   std::ofstream file("mandel_array.ppm",std::ios_base::binary);
   file<<"P5\n"<<w<<" "<<h<<"\n255\n";
   unsigned char *ca=new unsigned char[w*h]; mandel.read(ca);
   file.write((char *)ca,w*h);
   
   
   double total_flops=0;
   for (int j=0;j<h;j++)
      for (int i=0;i<w;i++) {
	     total_flops+=4+11*ca[i+j*w];
      }
   delete[] ca;
   printf("Mandelbrot %s flop count: %.3f GF = %.2f flops/pixel\n",
   		mandel_pretty?"pretty":"fast",
   		total_flops,total_flops/(w*h));
   return total_flops;
}


/***************** Main Driver ***************/
/* Make two copies of this variable, first raw, then as a string */
#define x2(x) x,#x

/* Just list out all our tests.  List your development tests first! */
int main() {
	
	istencil_run();
	stencil_run();
	stencil8_run();
	perf_wxh(istencil_sweep,"image stencil_sweep",sizeof(float)*5,4);
	
	printf("Testing basic overheads:\n");
	perf_wxh(x2(write1Dplain),sizeof(float),1);
	perf_wxh(x2(write1D),sizeof(float),1);
	perf_wxh(x2(write1Dfill),sizeof(float),1);
	perf_wxh(x2(write2Drows),sizeof(float),1);
	perf_wxh(x2(write2Dcols),sizeof(float),1);
	perf_wxh(x2(write2Drowsinter),sizeof(float),1); /* <- this is surprisingly expensive! */
	perf_wxh(x2(write2Dinter),sizeof(float),1);
	perf_wxh(x2(write2Dinterji),sizeof(float),1);

#if BENCH_ATOMICS
	printf("Testing atomic operations:\n");
	atomic_sweep(x2(atom_inc_i));
	for (int bit=6;bit>=0;bit--) {
		atomic_mask=1<<bit;
		printf("%6x&",atomic_mask);
		atomic_sweep(x2(atom_inc_i));
	}
	atomic_mask=0;
	for (int bit=6;bit>=0;bit--) {
		atomic_xor=1<<bit;
		printf("%6x^",atomic_xor);
		atomic_sweep(x2(atom_inc_i));
	}
	if (0) { // really detailed atomic testing
		for (atomic_xor=63;atomic_xor>=0;atomic_xor--) {
			printf("%6x^",atomic_xor);
			atomic_sweep(x2(atom_inc_i));
		}
	}
	atomic_xor=0;
	atomic_sweep(x2(atom_add_i));
	
	//atomic_sweep(x2(atom_add_f)); // <- crashes some CUDA backends!
#endif
	
	printf("Testing Volkov-type multiple outputs per kernel:\n");
	perf_wxh(x2(AK_ivector),2*sizeof(float),1);
	perf_wxh(x2(AK2_ivector),2*2*sizeof(float),2);
	perf_wxh(x2(AK3_ivector),3*2*sizeof(float),3);
	
	printf("Testing matrix transpose:\n");
	perf_wxh(x2(rmw1D),2*sizeof(float),1);
	perf_wxh(x2(rmw2Drows),2*sizeof(float),1);
	perf_wxh(x2(rmw2Dcols),2*sizeof(float),1);
	perf_wxh(x2(rmw2Dinter),2*sizeof(float),1);
	perf_wxh(x2(rmw2Dinterji),2*sizeof(float),1);
	perf_wxh(x2(transpose2Dinter),2*sizeof(float),1);
	
	perf_wxh(x2(naive_transpose),sizeof(float)*2,1);
	perf_wxh(x2(local_transpose),sizeof(float)*2,1);
	
	for (mandel_pretty=0;mandel_pretty<=1;mandel_pretty++) {
		double total_flops=mandelbrot();
		perf_wxh(x2(mandel_renderer),sizeof(float)*1,total_flops/(w*h));
	}
	model_alphabeta(x2(paperdemo),0,24);
#if BENCH_DOUBLE
	model_alphabeta(x2(poly1d),0,24);
#endif
	model_alphabeta(x2(poly1),0,24);
	model_alphabeta(x2(poly2),0,24);
	model_alphabeta(x2(poly3),0,24);
	model_alphabeta(x2(poly4),0,24);
	
	model_alphabeta(x2(AK_ivector),0,24);
	model_alphabeta(x2(AK2_ivector),0,24);
	model_alphabeta(x2(AK3_ivector),0,24);
	
	model_alphabeta(x2(K_fvector),0,24);
	
	model_alphabeta(x2(just_clFinish),0,24);
	model_alphabeta(x2(K_ivector),0,24);
	model_alphabeta(x2(K_fvector),0,24);
	model_alphabeta(x2(K1_ivector),0,24);
	model_alphabeta(x2(K2_ivector),0,24);
	model_alphabeta(x2(K3_ivector),0,24);
	model_alphabeta(x2(R_ivector),0,24);
	model_alphabeta(x2(R_fvector),0,24);
	model_alphabeta(x2(MR_ivector),0,24);
	model_alphabeta(x2(W_ivector),0,24);
	model_alphabeta(x2(W_fvector),0,24);
	model_alphabeta(x2(MW_ivector),0,24);
	model_alphabeta(x2(WR_ivector),0,24);
	model_alphabeta(x2(MRW_ivector),0,24);
	model_alphabeta(x2(HMRW_ivector),0,24);
	model_alphabeta(x2(WKR_ivector),0,24);
	
	for (int cache_enabled=1;cache_enabled>=0;cache_enabled--) {
		gpu_env::static_env().buffer_reuse_enabled=cache_enabled;
		printf("Buffer reuse %s:\n",cache_enabled?"enabled":"disabled");
		model_alphabeta(x2(A_ivector),0,24);
		model_alphabeta(x2(AK_ivector),0,24);
		model_alphabeta(x2(AK_fvector),0,24);
		model_alphabeta(x2(AWR_ivector),0,24);
		model_alphabeta(x2(AWKR_ivector),0,24);
		model_alphabeta(x2(allocate_cvector),0,26);
		model_alphabeta(x2(allocate_dvector),0,23);
	}
	gpu_env::static_env().buffer_reuse_enabled=true;
	
	measure_workgroup();


	return 0;
}
