/**
Test for EPGPU OpenCL library: 2D rendering of 3D particles

Dr. Orion Lawlor, lawlor@alaska.edu, 2011-04-27 (Public Domain)
*/
#include <iostream>
#include <stdlib.h>
#include "epgpu.h" /* Orion Lawlor's Expressive Programming for GPU library */
#include "epgpu.cpp" /* <- just compile it here, now.  No compiler/DLL hassles! */
#include "timingCL.h" /* benchmarking utilities */

GPU_ADD_STATIC_CODE(
"#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable\n"
"#pragma OPENCL EXTENSION cl_khr_global_int32_extended_atomics : enable\n"
,extension_atomic_inc_pragma);

GPU_STRUCT(vec3,
	float x;
	float y;
	float z;
,/* C++-only */
	vec3() {}
	vec3(float x_,float y_,float z_) { x=x_; y=y_; z=z_; }
	float dot(vec3 &other) const {return x*other.x+y*other.y+z*other.z;}
);

GPU_STRUCT(viewpoint3d,
	vec3 o; /* origin */
	vec3 x; /* dot with vectors to get x coordinate */
	vec3 y; /* dot with vectors to get y coordinate */
,/* C++-only */
);

int w,h; 
double start;
template <class T>
void endTest(const char *name,gpu_vector2d<T> &arr) {
	T v=arr(0,0); /* finish the damn kernel! */
	double elapsed=time_in_seconds()-start;
	std::cout<<name<<": "<<elapsed/(w*h)*1.0e9<<" ns per (v="<<v<<")\n";
}

int main(int argc,char *argv[]) {
	w=h=1024;
	if (argc>2) {w=atoi(argv[1]); h=atoi(argv[2]); }
	
	static GPU_KERNEL_2D(writeArray, (int w, __global<int *> dest),
		dest[i+w*j]=100000*j+1000*i+3;
	)
	static GPU_KERNEL_2D(maxArray, (int w, __global<int *> dest),
		atom_max(&dest[i+w*j],100000*j+1000*i+3);
	)
	static GPU_KERNEL_2D(max3DArray, (int w, __global<int *> dest, __constant<viewpoint3d *> vp),
		vec3 cur; cur.x=i-vp->o.x; cur.y=j-vp->o.y; cur.z=0.0;
		
		int dest_i=(int)(cur.x*vp->x.x+cur.y*vp->x.y);
		int dest_j=(int)(cur.x*vp->y.x+cur.y*vp->y.x);
		atom_max(&dest[dest_i+w*dest_j],100000*j+1000*i+3);
	)
	static GPU_KERNEL_2D(maxXORArray, (int w, __global<int *> dest),
		atom_max(&dest[(i^j)+w*j],100000*j+1000*i+3);
	)
	static GPU_KERNEL_2D(maxDivArray, (int w, __global<int *> dest),
		atom_max(&dest[(i/(1+j))+w*j],100000*j+1000*i+3);
	)
	static GPU_FILLKERNEL_2D(int, fillArray, (), 
		result=0;
	)
	static GPU_FILLKERNEL_2D(unsigned char, fillCArray, (), 
		result=0;
	)
	static GPU_KERNEL_2D(writeCArray, (int w, __global<unsigned char *> dest),
		dest[i+w*j]=16*j+i;
	)
	
	/* No char8 atomic operations; have to build them using int operations. 
	   This is surprisingly tricky, and performs poorly (5ns/pixel)
	   *iff* there are collisions.  Performance is OK (1.5ns/pixel)
	   without collisions.  Using "int" is simpler and faster, though.
	*/
	static GPU_KERNEL_2D(maxCArray, (int w, __global<unsigned char *> dest),
		unsigned int v=16*j+i; // our char value (to max)
		unsigned int off=i&3; // our byte offset within an int
		unsigned int shift=off*8; // our bit offset within int
		unsigned int mask=0xff;
		unsigned int mask_shift=mask<<shift;
		unsigned int v_shift=v<<shift;
		__global unsigned int *mem;
		//mem=&dest[((4*(i))%w)+w*j]; // spread out across machine
		mem=&dest[(i-off)+w*j]; // round down to nearest integer
		unsigned int old=*mem;
		unsigned int last_old;
		do {
			last_old=old;
			unsigned int nu=old;
			if ((old&mask_shift)<v_shift) { /* we're bigger! */
				nu=(v_shift)|(old&(~mask_shift));
			}
			old=atom_cmpxchg(mem,old,nu);
		} while (old!=last_old);
	)
	
	gpu_vector2d<unsigned char> carr(w,h);
	carr=fillCArray(); start=time_in_seconds();
	carr=writeCArray(w,carr);
	endTest("write byte",carr);
	
	carr=fillCArray(); start=time_in_seconds();
	carr=maxCArray(w,carr);
	endTest("max byte",carr);
	for (int j=0;j<16;j++) {
		for (int i=0;i<16;i++)
			std::cout<<std::hex<<(int)carr(i,j)<<std::dec<<" ";
		std::cout<<"\n";
	}
	
	gpu_vector2d<int> arr(w,h);
	arr=fillArray();
	start=time_in_seconds();
	arr=writeArray(w,arr);
	endTest("write int",arr);
	
	arr=fillArray();
	start=time_in_seconds();
	arr=maxArray(w,arr);
	endTest("max",arr);
	
	for (int j=0;j<h;j+=100) {
		for (int i=0;i<w;i+=100)
			std::cout<<arr(i,j)<<" ";
		std::cout<<"\n";
	}
	
	arr=fillArray();
	start=time_in_seconds();
	viewpoint3d vp;
	vp.o=vec3(0,0,0);
	vp.x=vec3(1,0,0);
	vp.y=vec3(0,1,0);
	arr=max3DArray(w,arr,vp);
	endTest("3D",arr);
	
	arr=fillArray();
	start=time_in_seconds();
	arr=maxXORArray(w,arr);
	endTest("xor",arr);
	
	arr=fillArray();
	start=time_in_seconds();
	arr=maxDivArray(w,arr);
	endTest("div",arr);
	
	
	return 0;
}
