/**
Breadth-first graph traversal on the GPU

Dr. Orion Lawlor, lawlor@alaska.edu, 2011-05-17 (Public Domain)
*/
#include <iostream>
#include <deque>
#include <stdlib.h>
#include "epgpu.h" /* Orion Lawlor's Expressive Programming for GPU library */
#include "epgpu.cpp" /* <- just compile it here, now.  No compiler/DLL hassles! */
#include "timingCL.h"

/*
static GPU_KERNEL_2D(writeArray, (int w, __global<int *> dest),
	dest[i+j*w]=interleave2D(i,j);
)
*/

/** CPU Naive std::deque version.
  Write into visitorder[i] the stage at which 
  a breadth-first search of this graph reaches node i. */
void traverseN(unsigned int nnode,
	const std::vector<unsigned int> &start,
	const std::vector<unsigned int> &adj,
	unsigned int firstnode,
	std::vector<unsigned int> &visitorder)
{
	for (unsigned int n=0;n<nnode;n++) visitorder[n]=0;
	std::deque<unsigned int> Q;
	Q.push_front(firstnode); 
	int order=1;
	while (!Q.empty()) {
		unsigned int n=Q.back(); Q.pop_back();
		if (visitorder[n]==0) { /* we haven't been here before */
			visitorder[n]=order++;
			for (unsigned int adjentry=start[n];adjentry<start[n+1];adjentry++)
			{
				// if (visitorder[adj[adjentry]]==0) // early exit (no real help)
					Q.push_front(adj[adjentry]);
			}
		}
	}
}

/** CPU pointer-based preallocated version.
  Write into visitorder[i] the stage at which 
  a breadth-first search of this graph reaches node i. */
void traverseP(unsigned int nnode,
	const unsigned int *start,
	const unsigned int *adj,
	unsigned int firstnode,
	unsigned int *queue,
	unsigned int *visitorder)
{
	for (unsigned int n=0;n<nnode;n++) visitorder[n]=0;
	unsigned int *queuefront=queue;
	unsigned int *queueback=queue;
	*queuefront++=firstnode;
	int order=1;
	while (queueback!=queuefront) {
		unsigned int n=*queueback++;
		if (visitorder[n]==0) { /* we haven't been here before */
			visitorder[n]=order++;
			for (unsigned int adjentry=start[n];adjentry<start[n+1];adjentry++)
				*queuefront++ = adj[adjentry];
		}
	}
}


GPU_ADD_STATIC_CODE(
"#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable\n"
"#pragma OPENCL EXTENSION cl_khr_global_int32_extended_atomics : enable\n"
,extension_atomic_inc_pragma);

GPU_KERNEL(traverseG,(int nnode,
	__global<unsigned int *> start,__global<unsigned int *> adj,
	__global<unsigned int *> queue,volatile __global<unsigned int *> counters,
	__global<unsigned int *> visitorder),

{
	// Atomically claim the next queue entry
	unsigned int front;
	front=counters[0];
	if (front>=counters[1]) return; // we passed the end
	unsigned int worked=atom_cmpxchg(&counters[0],front,front+1);
	if (worked==front) { // we won the race for the counter
	
		unsigned int n=queue[front];
		if (n>=nnode) return; // fell off the end?
		if (visitorder[n]!=0) return; // somebody else has already been here
		visitorder[n]=atom_inc(&counters[2]); /* mark write as in-progress */

		unsigned int beg=start[n];
		unsigned int end=start[n+1];
		unsigned int len=end-beg;
		
		// Count up our un-traversed neighbors
		unsigned int nnbors=0;
		for (unsigned int i=0;i<len;i++)
			if (visitorder[adj[beg+i]]==0) // not traversed yet
				nnbors++;

		// Write those neighbors into the global shared queue
		unsigned int queuestart=atom_add(&counters[1],nnbors);
		unsigned int j=0;
		for (unsigned int i=0;i<len;i++)
			if (visitorder[adj[beg+i]]==0) {
				queue[queuestart+j] = adj[beg+i];
				j++;
			}
		// due to race condition, may get here...
		for (;j<nnbors;j++) queue[queuestart+j]=nnode; // mark as invalid
	}
}
)

int main(int argc,char *argv[]) {
  for (unsigned int nnode=10;nnode<=10000;nnode*=10) {
	
	std::vector<unsigned int> start; /* indexed by node: first entry in adj */
	std::vector<unsigned int> adj; /* adjacent links */
	for (unsigned int node=0;node<nnode;node++) {
		start.push_back(adj.size());
		unsigned int nnbor=rand()%10; // fixme: add power law here
		for (unsigned int nbor=0;nbor<nnbor;nbor++) {
			unsigned int n;
			do { n=rand()%nnode; } while (n==node); // avoid self-adjacency
			adj.push_back(n);
		}
	}
	start.push_back(adj.size());
	
	// CPU-side naive std::queue traversal:
	unsigned int firstnode=0;
	std::vector<unsigned int> visitorderN(nnode,0);
	
	double startT=time_in_seconds();
	traverseN(nnode,start,adj,firstnode,visitorderN);
	double nsperN=(time_in_seconds()-startT)*1.0e9/adj.size();
	
	
	// CPU-side pointer-based preallocated traversal:
	std::vector<unsigned int> visitorderP(nnode,0);
	std::vector<unsigned int> queue(adj.size()*2);
	startT=time_in_seconds();
	traverseP(nnode,&start[0],&adj[0],firstnode,&queue[0],&visitorderP[0]);
	double nsperP=(time_in_seconds()-startT)*1.0e9/adj.size();
	
	
	// GPU-side traversal:
	gpu_vector<unsigned int> startG(start);
	gpu_vector<unsigned int> adjG(adj);
	gpu_vector<unsigned int> visitG(std::vector<unsigned int>(nnode,0));
	gpu_vector<unsigned int> queueG(adj.size()*2);
	queueG.write(&firstnode,0,1);
	
	/* These are atomically incremented counters on the GPU */
	std::vector<unsigned int> counters(3,0);
#define CTR_BACK 0
#define CTR_FRONT 1
#define CTR_VISIT 2
	counters[0]=0; /* queue_back */
	counters[1]=1; /* queue_front */
	counters[2]=1; /* visit counter */
	gpu_vector<unsigned int> countersG(counters);

	int nprint=0;
	do {
		traverseG(nnode,startG,adjG,queueG,countersG,visitG).run(1024*16);
		if((++nprint%1000)==0)
			printf("GPU still working: %d %d %d\n",
				countersG[0],countersG[1],countersG[2]);
	} while (countersG[0]<countersG[1]);
	double nsperG=(time_in_seconds()-startT)*1.0e9/adj.size();
	std::vector<unsigned int> visitorderG(nnode);
	visitG.read(visitorderG);
	
	
	// Sanity checks:
	int errorcount=0;
	for (unsigned int node=0;node<nnode;node++) 
	{
		if (visitorderN[node]!=visitorderP[node])
			printf("	Node %3d: %d N vs %d P\n",node,
				visitorderN[node],visitorderP[node]);
		if (visitorderN[node]!=visitorderG[node]) {
			errorcount++;
			if (errorcount<10)
			printf("	Node %3d: %d N vs %d G\n",node,
				visitorderN[node],visitorderG[node]);
		}
	}
	if (errorcount>0) printf("	Total of %d differences\n",errorcount);
	if (nnode<100) {
		printf("Visit order:\n");
		for (unsigned int node=0;node<nnode;node++) 
			printf("%3d ",visitorderN[node]);
		printf("\n");
	}
	
	printf("For %8d-node graph: %.2f ns naive, %.2f ns pointer, %.2f ns GPU\n",nnode,
		nsperN,nsperP,nsperG); fflush(stdout);

  }
	
	return 0;
}
