/**
Render 3D particles into a 2D framebuffer.

Dr. Orion Lawlor, lawlor@alaska.edu, 2012-01-04 (Public Domain)
*/
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include "epgpu.h" /* Orion Lawlor's Expressive Programming for GPU library */
#include "epgpu.cpp" /* <- just compile it here, now.  No compiler/DLL hassles! */
#include "timingCL.h" /* benchmarking utilities */

GPU_ADD_STATIC_CODE(
"#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable\n"
"#pragma OPENCL EXTENSION cl_khr_global_int32_extended_atomics : enable\n"
,extension_atomic_inc_pragma);

GPU_STRUCT(vec3,
	float x;
	float y;
	float z;
,/* C++-only */
	vec3() {}
	vec3(float x_,float y_,float z_) { x=x_; y=y_; z=z_; }
	float dot(vec3 &other) const {return x*other.x+y*other.y+z*other.z;}
	friend vec3 operator*(float s,const vec3 &v) {
		return vec3(s*v.x,s*v.y,s*v.z);
	}
);

/* Random float between 0 and 1 */
float randFloat(void) {
        float f=rand();
        return f/RAND_MAX;
}

GPU_STRUCT(particle,
	vec3 loc;
	int value;
,/* C++-only */
	particle() {loc.x=randFloat(); loc.y=randFloat(); loc.z=randFloat(); value=(int)(255.99*randFloat());}
);

GPU_STRUCT(viewpoint3d,
	vec3 o; /* origin */
	vec3 x; /* dot with vectors to get x coordinate */
	vec3 y; /* dot with vectors to get y coordinate */
,/* C++-only */
);

int w,h; 
double start;
template <class T>
void endTest(const char *name,gpu_vector2d<T> &arr) {
	T v=arr(0,0); /* finish the darn kernel! */
	double elapsed=time_in_seconds()-start;
	std::cout<<name<<": "<<elapsed/(w*h)*1.0e9<<" ns per (v="<<v<<")\n";
}

int main(int argc,char *argv[]) {
	w=h=1024;
	if (argc>2) {w=atoi(argv[1]); h=atoi(argv[2]); }
	
	/* The framebuffer is integer, so we can do atom_max blending. 
	  (there isn't an atom_max for unsigned char data types.) */
	static GPU_FILLKERNEL_2D(int, zeroArray, (),
		result=0;
	)
	
	static GPU_KERNEL(max3DArray, (int w, int h, 
			__global<particle *> particles, int n_particles,
			__global<int *> framebuffer, __constant<viewpoint3d *> vp),
	 {
		vec3 cur=particles[i].loc; 
		cur.x-=vp->o.x; cur.y-=vp->o.y; cur.z-=vp->o.z;
		int dest_i=(int)(cur.x*vp->x.x+cur.y*vp->x.y+cur.z*vp->x.z);
		int dest_j=(int)(cur.x*vp->y.x+cur.y*vp->y.y+cur.z*vp->y.z);
		
		if (dest_i>=0 && dest_i<w && dest_j>=0 && dest_j<h && i<n_particles)
			atom_max(&framebuffer[dest_i+w*dest_j],particles[i].value);
	 }
	)
	
	std::vector<particle> cpu_particles;
	for (int i=0;i<10*1000*1000;i++) cpu_particles.push_back(particle());
	gpu_vector<particle> particles(cpu_particles);
	
	gpu_vector2d<int> framebuffer(w,h);
	
	for (int repeat=0;repeat<3;repeat++) {
		start=time_in_seconds();
		framebuffer=zeroArray();
		viewpoint3d vp;
		// Plain ortho view
		vp.o=vec3(0,0,0);
		vp.x=vec3(w,0,0);
		vp.y=vec3(0,h,0);
		
		if (1) { // cool skewed view
			vp.o=vec3(-0.3,-0.2,0);
			vp.x=w*vec3(0.5,0.3,-0.2);
			vp.y=h*vec3(-0.1,0.7,0.2);
		}
		
		max3DArray(w,h, 
			particles,particles.size(),
			framebuffer,vp).run(particles.size());
		endTest("3D",framebuffer);
	}
	
    // Write rendered data to output file
	std::ofstream file("particle_render.ppm");
	file<<"P5\n"<<w<<" "<<h<<"\n255\n";
	int *ia=new int[w*h];
	framebuffer.read(ia);
	unsigned char *ca=new unsigned char[w*h];
	for (int i=0;i<w*h;i++) ca[i]=ia[i];
	file.write((char *)ca,w*h);
	
	return 0;
}
